/*
 *  Copyright (C) 2009 Nicolas Vion <nico@picapo.net>
 *
 *   This file is part of Fluxy.
 *
 *   Fluxy is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Fluxy is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Fluxy; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#define _AUTODUMP_CC

#include "autodump.hh"

#include "log.hh"
#include "users.hh"
#include "stringutils.hh"
#include "stats.hh"

void AutoDump::main() {
	for (;;) {	
		int delay;
		std::string target;
		bool enabled;
		settings.get(enabled, target, delay);

		if (enabled) {
			sleep(delay);
			dump(target);
		}
		else {
			sleep(60);
		}
	}
}

bool AutoDump::force() {
	int delay;
	std::string target;
	bool enabled;
	settings.get(enabled, target, delay);
	if (!enabled or target == "")
		return false;

	return dump(target);
}

bool AutoDump::dump(std::string const target) {
	mutex.lock();
	log.msg(LOG_NOTICE, "Start autodump into '" + target + "'.");

	bool result = users.dump(target);

	std::string message = (result) ? "Autodump successful." : "Could not dump data to '" + target + "'.";
	log.msg((result) ? LOG_NOTICE : LOG_ERR, message);
	report.set(result, message);
	mutex.unlock();
	return result;
}

void AutoDumpSettings::set(bool const _enabled, std::string const _target, int const _delay) {
	mutex.lock();
	enabled = _enabled;
	target = _target;
	delay = _delay;
	log.msg(LOG_NOTICE, "Autodump settings: " + std::string(enabled ? "enabled" : "disabled") + ", \"" + target + "\", " + StringUtils::to_string(delay));
	mutex.unlock();
}

void AutoDumpSettings::get(bool &_enabled, std::string &_target, int &_delay) {
	mutex.lock();
	_enabled = enabled;
	_target = target;
	_delay = delay;
	mutex.unlock();
}

void AutoDumpSettings::serialize_php(std::stringstream &out) {
	mutex.lock();
	out << "a:3:{";
	out << "s:7:\"enabled\";b:" << (enabled ? "1" : "0") << ";";
	out << "s:6:\"target\";s:" << target.size() << ":\"" << target << "\";";
	out << "s:5:\"delay\";i:" << delay << ";";
	out << "}";
	mutex.unlock();
}

bool AutoDump::parse_query(ClientResult &result, std::string const cmd_prefix, WordsParser *parser, OutputType const mode) {
	//!stats
	//!	Get stats about last autodump
	if (parser->current == "stats") {
		stats.inc(cmd_prefix + "stats");
		if (mode == TEXT) {
			result.type = TEXT;
			report.show(result.data);
		}
		else {
			result.type = PHP_SERIALIZE;
			report.serialize_php(result.data);
		}
		result.send();
		return true;
	}

	//!get
	//!	Get autodump settings
	if (parser->current == "get") {
		stats.inc(cmd_prefix + "get");
		result.type = PHP_SERIALIZE;
		settings.serialize_php(result.data);
		result.send();
		return true;
	}

	//!set <target> <delay>
	//!	Set autodump settings
	if (parser->current == "set") {
		stats.inc(cmd_prefix + "set");

		std::string target = parser->next(true);
		time_t delay = StringUtils::to_time(parser->next());
		settings.set(true, target, (delay != 0) ? delay : 3600);
		result.send();
		return true;
	}

	//!disable
	//!	Disable autodump
	if (parser->current == "disable") {
		stats.inc(cmd_prefix + "disable");
		settings.set(false, "", 60);
		result.send();
		return true;
	}

	//!force
	//!	Force autodump
	if (parser->current == "force") {
		stats.inc(cmd_prefix + "force");
		if (!force())
			result.error("Could not dump data.");
		result.send();
		return true;
	}

	//!help
	//!	Show commands list
	if (parser->current == "help") {
		stats.inc("misc");
		result.data << HELP_AUTODUMP;
		result.send();
		return true;
	}
	return false;
}

void AutoDumpReport::set(bool const _result, std::string const _message) {
	mutex.lock();
	setted = true;
	at = time(NULL);
	result = _result;
	message = _message;
	mutex.unlock();
}

void AutoDumpReport::serialize_php(std::stringstream &out) {
	mutex.lock();
	if (!setted) {
		out << "b:0;";
	}
	else {
		out << "a:3:{";
		out << "s:5:\"dtime\";i:" << time(NULL) - at << ";";
		out << "s:6:\"result\";b:" << ((result) ? 1 : 0) << ";";
		out << "s:7:\"message\";s:" << message.size() << ":\""  << message << "\";";
		out << "}";
	}
	mutex.unlock();
}

void AutoDumpReport::show(std::stringstream &out, std::string const prefix) {
	mutex.lock();
	if (setted) {
		out << "STAT " << prefix << "dtime " << time(NULL) - at << std::endl;
		out << "STAT " << prefix << "result " << ((result) ? "TRUE" : "FALSE") << std::endl;
		out << "STAT " << prefix << "message " << message << std::endl;
	}
	mutex.unlock();
}

AutoDumpReport::AutoDumpReport() : setted(false), at(0), result(true) {
}
