/*
 *  Copyright (C) 2009 Nicolas Vion <nico@picapo.net>
 *
 *   This file is part of Fluxy.
 *
 *   Fluxy is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Fluxy is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Fluxy; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#define _USER_CC

#include "user.hh"

#include <algorithm>
#include <cstring>
#include <iostream>

#include "help.hh"
#include "stats.hh"
#include "addstats.hh"
#include "stringutils.hh"
#include "recycler.hh"
#include "replicator.hh"
#include "autoid.hh"
#include "fetcher.hh"
#include "log.hh"

#define MESSAGE_SET_ID(message, value) \
	message->id = value; 

#define MESSAGE_GET_ID(message) \
	message->id

#define MESSAGE_SET_DATE(message, value) \
	message->date = value;

#define MESSAGE_GET_DATE(message) \
	message->date

#define MESSAGE_SET_TYPE(message, value) \
	message->type = value;

#define MESSAGE_GET_TYPE(message) \
	message->type

#define MESSAGE_SET_FLAGS(message, value) \
	message->flags = value;

#define MESSAGE_GET_FLAGS(message) \
	message->flags

#define MESSAGE_SET_VERSION(message, value) \
	message->version = ((message->version & 0x80) | (value & 0x7f));

#define MESSAGE_GET_VERSION(message) \
	(message->version & 0x7f)

#define MESSAGE_SET_SEEN(message, value) \
	message->version = ((message->version & 0x7f) | (value ? 0x80 : 0));

#define MESSAGE_GET_SEEN(message) \
	((message->version & 0x80) == 0x80)

#define MESSAGE_SET_BODY(message, str) \
	message->body_len = str.size(); \
	memcpy(&message->body, str.data(), message->body_len);

#define MESSAGE_GET_BODY(message) \
	std::string(&message->body, message->body_len)

#define MESSAGE_GET_BODY_LEN(message) \
	message->body_len

#define MESSAGE_SIZE(body_len) \
	(sizeof(UserMessageStorage) - sizeof(char) + body_len)

#define MESSAGE_SET(message, values) \
	MESSAGE_SET_ID(message, values.id) \
	MESSAGE_SET_DATE(message, values.date) \
	MESSAGE_SET_TYPE(message, values.type) \
	MESSAGE_SET_FLAGS(message, values.flags) \
	MESSAGE_SET_VERSION(message, values.version) \
	MESSAGE_SET_SEEN(message, values.seen) \
	MESSAGE_SET_BODY(message, values.body)

#define MESSAGE_GET(message, values) \
	values.id = MESSAGE_GET_ID(message); \
	values.date = MESSAGE_GET_DATE(message); \
	values.type = MESSAGE_GET_TYPE(message); \
	values.flags = MESSAGE_GET_FLAGS(message); \
	values.version = MESSAGE_GET_VERSION(message); \
	values.seen = MESSAGE_GET_SEEN(message); \
	values.body = MESSAGE_GET_BODY(message);


void UserMessageSearch::parse(WordsParser *parser) {
	id = StringUtils::to_uint(parser->next());
	precise_type = false;
	parser->next();
	if (parser->current == ":") {
		precise_type = true;
		type = StringUtils::to_uint(parser->next());
		parser->next();
	}
}

void UserMessageSearch::output(std::stringstream &out) {
	out << id;
	if (precise_type) 
		out << ":" << (int) type;
}

int UserMessagesIndex::make(void *buffer, int const count, time_t const until) {
	size_t size_from_begining = 0;
	char *cursor = (char *) buffer;

	for (int i = 0; i < count; i++) {
		UserMessageStorage *message = (UserMessageStorage *) cursor;

		if (until != 0 and message->date < until)
			return i;

		size_t body_len = message->body_len;
		size_t size = MESSAGE_SIZE(body_len);
		size_from_begining += size;

		items[i].message = message;
		items[i].body_len = body_len;
		items[i].size = size;
		items[i].size_from_begining = size_from_begining;

		cursor += size;
	}
	return count;
}

int UserMessagesIndex::find(int const count, UserMessageSearch const search) {
	for (int i = 0; i < count; i++) {
		UserMessageStorage *message = items[i].message;
		if (message->id == search.id and (!search.precise_type or message->type == search.type))
			return i;
	}
	return -1;
}

size_t UserMessagesIndex::size_from_begining(int const count) {
	return (count == 0) ? 0 : items[count - 1].size_from_begining;
}

UserMessages::UserMessages() : buffer_count(0), buffer(NULL) {
}

void UserMessages::clear() {
	buffer_count = 0;
	alloc(0);
}

bool UserMessages::alloc(size_t const needed_size) {
	if (needed_size == 0) {
		if (buffer != NULL) {
			recycler.lock();
			recycler.free(buffer, buffer_scale);
			recycler.unlock();
			buffer = NULL;
		}
		return true;
	}

	BufferScale needed_scale;
	size_t size;
	if (!Recycler::calc_scale(needed_size, needed_scale, size)) {
		return false;
	}

	if (buffer == NULL) {
		recycler.lock();
		buffer = recycler.alloc(needed_scale);
		recycler.unlock();
		buffer_scale = needed_scale;
	}
	else if (needed_scale != buffer_scale) {
		recycler.lock();
		buffer = recycler.realloc(buffer, buffer_scale, needed_scale);
		recycler.unlock();
		buffer_scale = needed_scale;
	}

	return true;
}

void UserMessages::debug_buffer() {
	size_t size = (buffer == NULL) ? 0 : (BUFFER_SCALE_SIZE(buffer_scale));
	uint8_t *data = (uint8_t *) buffer;
	std::cout << "Memory:" << std::endl;
	for (size_t i = 0; i < size; i++) {
		std::cout << (int) data[i] << " ";
	}
	std::cout << std::endl;
}

void UserMessages::cleanup() {
	UserMessagesIndex index;
	buffer_count = index.make(buffer, buffer_count, time(NULL) - user_messages_expiration);
	alloc(index.size_from_begining(buffer_count));
}

bool UserMessages::add(UserMessageValues const values) {
	//stats
	addstats.inc(values.type);

	//delete old messages if already user_messages_size messages
	buffer_count = std::min((uint8_t) (user_messages_size - 1), buffer_count);

	UserMessagesIndex index;
	buffer_count = index.make(buffer, buffer_count, time(NULL) - user_messages_expiration);

	size_t size = index.size_from_begining(buffer_count);

	size_t msg_size = MESSAGE_SIZE(values.body.size());

	if (!alloc(msg_size + size)) {
		return false;
	}

	//translate old messages
	if (size != 0) {
		memmove(buffer + msg_size, buffer, size);
	}

	//set new message content
	UserMessageStorage *message = (UserMessageStorage *) buffer;
	MESSAGE_SET(message, values)
	buffer_count++;
	return true;
}

bool UserMessages::del(UserMessageSearch const search) {
	UserMessagesIndex index;
	buffer_count = index.make(buffer, buffer_count, time(NULL) - user_messages_expiration);

	int idx = index.find(buffer_count, search);
	if (idx < 0)
		return false;

	UserMessagesIndexItem item = index.items[idx];
	memmove(item.message, (char *) item.message + item.size, index.items[buffer_count - 1].size_from_begining - item.size_from_begining);
	buffer_count--;
	return true;
}

bool UserMessages::update(UserMessageSearch const search, UserMessageFilter const filter, UserMessageValues const values) {
	UserMessagesIndex index;
	buffer_count = index.make(buffer, buffer_count, time(NULL) - user_messages_expiration);

	int idx = index.find(buffer_count, search);
	if (idx < 0)
		return false;

	UserMessagesIndexItem item = index.items[idx];

	if (filter.date) {
		MESSAGE_SET_DATE(item.message, values.date)
	}

	if (filter.type) {
	        MESSAGE_SET_TYPE(item.message, values.type)
	}

	if (filter.flags) {
	        MESSAGE_SET_FLAGS(item.message, values.flags)
	}

	if (filter.version) {
	        MESSAGE_SET_VERSION(item.message, values.version)
	}

	if (filter.seen) {
	        MESSAGE_SET_SEEN(item.message, values.seen)
	}

	if (filter.body) {
		//realloc if necessary
		size_t new_size = MESSAGE_SIZE(values.body.size());
		if (new_size > item.size) {
			char *oldbuf = buffer;
			alloc(new_size - item.size + index.items[buffer_count - 1].size_from_begining);

			//update pointer if buffer has been moved
			if (buffer != oldbuf) {
				item.message = (UserMessageStorage*) (buffer + item.size_from_begining - item.size);
			}
		}

		if (new_size != item.size) {
			memmove((char *) item.message + new_size, (char *) item.message + item.size, index.items[buffer_count - 1].size_from_begining - item.size_from_begining);
		}
		MESSAGE_SET_BODY(item.message, values.body)
	}
	return true;
}

bool UserMessages::seen(bool const all, UserMessageSearch const last) {
	UserMessagesIndex index;
	buffer_count = index.make(buffer, buffer_count, time(NULL) - user_messages_expiration);

	int idx = (all) ? 0 : index.find(buffer_count, last);
	if (idx < 0)
		return false;

	for (int i = idx; i < buffer_count; i++) {
	        MESSAGE_SET_SEEN(index.items[i].message, true)
	}
	return true;
}

int UserMessages::count(bool const unseen_only) {
	UserMessagesIndex index;
	buffer_count = index.make(buffer, buffer_count, time(NULL) - user_messages_expiration);

	if (!unseen_only) 
		return buffer_count;

	int result = 0;
	for (int i = 0; i < buffer_count; i++)
		if (!MESSAGE_GET_SEEN(index.items[i].message))
			result++;
	return result;
}

void UserMessages::serialize_php(std::stringstream &out) {
	UserMessagesIndex index;
	buffer_count = index.make(buffer, buffer_count, time(NULL) - user_messages_expiration);

	out << "a:" << (int) buffer_count << ":{";
	for (int i = 0; i < buffer_count; i++ ) {
		UserMessageStorage *message = index.items[i].message;
		out << "i:" << i << ";a:7:{";
		out << "s:2:\"id\";i:" << MESSAGE_GET_ID(message) << ";";
		out << "s:4:\"date\";i:" << MESSAGE_GET_DATE(message) << ";";
		out << "s:4:\"type\";i:" << (int) MESSAGE_GET_TYPE(message) << ";";
		out << "s:5:\"flags\";i:" << MESSAGE_GET_FLAGS(message) << ";";
		out << "s:7:\"version\";i:" << (int) MESSAGE_GET_VERSION(message) << ";";
		out << "s:4:\"seen\";b:" << (MESSAGE_GET_SEEN(message) ? "1" : "0") << ";";
		out << "s:4:\"body\";s:" << MESSAGE_GET_BODY_LEN(message) << ":\"" << MESSAGE_GET_BODY(message) << "\";";
		out << "}";
	}
	out << "}";
}

void UserMessages::fetch(UserId const user_id, Fetcher &fetcher, UserMessageFlags const mask, UserMessageFlags const value, time_t const until, time_t const since) {
	UserMessagesIndex index;
	int msgs = index.make(buffer, buffer_count, until);

	UserMessageValues values;
	for (int i = 0; i < msgs; i++) {
		UserMessageStorage *message = index.items[i].message;
		if (since == 0 or message->date <= since) {
			if (mask == 0 or (message->flags & mask) == value) {
				MESSAGE_GET(message, values);
				fetcher.add(user_id, values);
			}
		}
	}
}

UserScore UserMessages::score() {
	if (buffer_count < user_messages_size) {
		return buffer_count - user_messages_size;
	}

	UserMessagesIndex index;
	index.make(buffer, buffer_count);
	time_t delta_t = index.items[0].message->date - index.items[user_messages_size - 1].message->date;
	return (delta_t == 0 ? 3600 * user_messages_size * 2 : 3600 * user_messages_size / delta_t);
}

void UserMessages::dump(FILE *f) {
	if (buffer) {
		UserMessagesIndex index;
		buffer_count = index.make(buffer, buffer_count, time(NULL) - user_messages_expiration);
		size_t size = index.size_from_begining(buffer_count);
		size_t buffer_size = BUFFER_SCALE_SIZE(buffer_scale);
		
		//Sanitary check
		if (size > buffer_size) {
			log.msg(LOG_ERR, "Internal error in " __FILE__ " line " + StringUtils::to_string(__LINE__), 1);
			size = buffer_size;
		}

		fwrite(&size, 1, sizeof(size), f);
		fwrite(buffer, 1, size, f);
	}
}

bool UserMessages::parse_query(ClientResult &result, std::string const cmd_prefix, WordsParser *parser, std::stringstream &replication_query) {
	//!messages add <id>|auto <date>|now <type> <flags> <version> <body>
	//!	Add new message
	if (parser->current == "add") {
		stats.inc(cmd_prefix + "add");
		parser->next();

		UserMessageValues values;

		//id
		values.id = 0;
		bool autoid = false;
		if (parser->current == "autoid") {
			parser->next();
			values.id = StringUtils::to_uint(parser->current);
			auto_id.set(values.id);
			autoid = true;
		}
		else if (parser->current == "auto") {
			values.id = auto_id.inc();
			autoid = true;
		}
		else {
			values.id = StringUtils::to_uint(parser->current);
		}
		parser->next();

		//date
		values.date = (parser->current == "now") ? time(NULL) : StringUtils::to_uint(parser->current);

		//type
		values.type = StringUtils::to_uint(parser->next());

		//flags
		values.flags = StringUtils::to_uint(parser->next());

		//version
		values.version = StringUtils::to_uint(parser->next());

		//seen
		values.seen = false;

		//body
		parser->next_char();
		values.body = parser->until_end();

		if (!add(values))
			return false;

		//replication
		if (replicator.opened and !result.replicated) {
			replication_query << "add ";
			if (autoid)
				replication_query << "autoid ";
			replication_query << values.id << " " << values.date << " " << (int) values.type << " " << values.flags << " " << (int) values.version << " " << values.body;
			replicator.add(replication_query);
		}

		result.type = PHP_SERIALIZE;
		result.data << "a:2:{s:2:\"id\";i:" << values.id << ";s:4:\"date\";i:" << values.date << ";}";
		return true;
	}

	//!messages delete <id>[:<type>]
	//!	Delete message
	if (parser->current == "delete") {
		stats.inc(cmd_prefix + "delete");

		UserMessageSearch search;
		search.parse(parser);

		if (!del(search))
			return false;

		//replication
		if (replicator.opened and !result.replicated) {
			replication_query << "delete ";
			search.output(replication_query);
			replicator.add(replication_query);
		}
		return true;
	}

	//!messages update <id>[:<type>] [date <date>] [type <type>] [flags <flags>] [version <version>] [seen <1/0>] [body <body>]
	//!	Update message
	if (parser->current == "update") {
		stats.inc(cmd_prefix + "update");

		UserMessageSearch search;
		UserMessageValues values;
		UserMessageFilter filter;

		//search
		search.parse(parser);

		//date
		filter.date = false;
		if (parser->current == "date") {
			filter.date = true;
			values.date = StringUtils::to_uint(parser->next());
			parser->next();
		}

		//type
		filter.type = false;
		if (parser->current == "type") {
			filter.type = true;
			values.type = StringUtils::to_uint(parser->next());
			parser->next();
		}

		//flags
		filter.flags = false;
		if (parser->current == "flags") {
			filter.flags = true;
			values.flags = StringUtils::to_uint(parser->next());
			parser->next();
		}

		//version
		filter.version = false;
		if (parser->current == "version") {
			filter.version = true;
			values.version = StringUtils::to_uint(parser->next());
			parser->next();
		}

		//seen
		filter.seen = false;
		if (parser->current == "seen") {
			filter.seen = true;
			values.seen = (parser->next() != "0");
			parser->next();
		}

		//body
		filter.body = false;
		if (parser->current == "body") {
			filter.body = true;
			parser->next_char();
			values.body = parser->until_end();
		}

		if (!update(search, filter, values))
			return false;

		//replication
		if (replicator.opened and !result.replicated) {
			replication_query << "update ";
			search.output(replication_query);
			if (filter.date) replication_query << " date " << values.date;
			if (filter.type) replication_query << " type " << (int) values.type;
			if (filter.flags) replication_query << " flags " << values.flags;
			if (filter.version) replication_query << " version " << (int) values.version;
			if (filter.seen) replication_query << " seen " << (values.seen ? "1" : "0");
			if (filter.body) replication_query << " body " << values.body;
			replicator.add(replication_query);
		}
		return true;
	}

	//!messages get
	//!	Get messages
	if (parser->current == "get") {
		stats.inc(cmd_prefix + "get");
		result.type = PHP_SERIALIZE;
		serialize_php(result.data);
		return true;
	}

	//!messages seen all|last <id>[:<type>]
	//!	Mark messages as seen
	if (parser->current == "seen") {
		stats.inc(cmd_prefix + "seen");
		parser->next();

		bool all = false;
		UserMessageSearch last;

		if (parser->current == "all")
			all = true;
		else if (parser->current == "last")
			last.parse(parser);
		else
			return false;
		

		if (!seen(all, last))
			return false;

		//replication
		if (replicator.opened and !result.replicated) {
			replication_query << "seen";
			if (all) {
				replication_query << " all";
			} 
			else {
				replication_query << " last ";
				last.output(replication_query);
			}
			replicator.add(replication_query);
		}
		return true;
	}

	//!messages count [unseen]
	//!	Count messages
	if (parser->current == "count") {
		stats.inc(cmd_prefix + "count");
		bool unseen_only = (parser->next() == "unseen");

		result.type = PHP_SERIALIZE;
		result.data << "i:" << count(unseen_only) << ";";
		return true;
	}

	//!messages clear
	//!	Clear messages
	if (parser->current == "clear") {
		stats.inc(cmd_prefix + "clear");
		clear();

		//replication
		if (replicator.opened and !result.replicated) {
			replication_query << "clear";
			replicator.add(replication_query);
		}
		return true;
	}
	return false;
}

PMutex *User::lock() {
	PMutex *mutex = &user_locks.lock[id & 0x3FF];
	mutex->lock();
	return mutex;
}

void User::cleanup() {
	messages.cleanup();
}

void User::show(std::stringstream &out) {
	out << "Id: " << id << std::endl;
}

void User::serialize_php(std::stringstream &out) {
	out << "a:3:{";
	out << "s:2:\"id\";i:" << id << ";";
	out << "s:5:\"flags\";i:" << flags << ";";
	out << "s:8:\"messages\";";
	messages.serialize_php(out);
	out << "}";
}

bool User::parse_query(ClientResult &result, std::string const cmd_prefix, WordsParser *parser) {
	std::stringstream replication_query;
	if (replicator.opened and !result.replicated) {
		replication_query << "#user *" << id << " messages ";
	}

	if (parser->current == "messages") {
		parser->next();
		return messages.parse_query(result, cmd_prefix + "messages::", parser, replication_query); 
	}

	//!get
	//!	Get user data
	if (parser->current == "get") {
		stats.inc(cmd_prefix + "get");
		result.type = PHP_SERIALIZE;
		serialize_php(result.data);
		return true;
	}

	//!show
	//!	Show user data
	if (parser->current == "show") {
		stats.inc(cmd_prefix + "show");
		show(result.data);
		return true;
	}

	//!delete
	//!	Delete user
	if (parser->current == "delete") {
		stats.inc(cmd_prefix + "delete");
		clear();
		deleted = true;

		//replication
		if (replicator.opened and !result.replicated) {
			replication_query << "delete";
			replicator.add(replication_query);
		}
		return true;
	}

	//!clear
	//!	Clear user
	if (parser->current == "clear") {
		stats.inc(cmd_prefix + "clear");
		clear();

		//replication
		if (replicator.opened and !result.replicated) {
			replication_query << "clear";
			replicator.add(replication_query);
		}
		return true;
	}

	//!help
	//!	Show commands list
	if (parser->current == "help") {
		stats.inc("misc");
		result.data << HELP_USER;
		return true;
	}
	return false;
}

void User::clear() {
	flags = 0;
	messages.clear();
}

UserScore User::score() {
	return messages.score();
}

void User::dump(FILE *f) {
	fwrite(this, 1, sizeof(User), f);
	messages.dump(f);
}

//TODO: stop restoring process if SANITARY_CHECK fault
#define SANITARY_CHECK(readen, size) \
	if (readen != size) { \
		log.msg(LOG_ERR, "Internal error in " __FILE__ " line " + StringUtils::to_string(__LINE__), 1); \
		return; \
	}

User::User(FILE *f) {
	size_t readen;
	readen = fread(this, 1, sizeof(User), f);
	SANITARY_CHECK(readen, sizeof(User));

	if (messages.buffer) {
		size_t buffer_size = BUFFER_SCALE_SIZE(messages.buffer_scale);
		size_t size;
		readen = fread(&size, 1, sizeof(size), f);
		SANITARY_CHECK(readen, sizeof(size));

		if (size > buffer_size) {
			log.msg(LOG_ERR, "Internal error in " __FILE__ " line " + StringUtils::to_string(__LINE__), 1);
			messages.buffer_count = 0;
			return;
		}

		messages.buffer = (char *) malloc(buffer_size);
		readen = fread(messages.buffer, 1, size, f);
		SANITARY_CHECK(readen, size);
	}
}

User::User(UserId const _id) : id(_id), flags(0), deleted(false) {
}

void UsersReport::add(User &user) {
	count++;
	messages_count += user.messages.buffer_count;
	msgstats[user.messages.buffer_count]++;
	if (user.deleted) {
		deleted_count++;
	}

	if (user.messages.buffer == NULL) {
		empty++;
	}
	else {
		int idx = user.messages.buffer_scale - BUFFER_SCALE_MIN;
		if (idx >= 0 and idx < MEMSTATS_COUNT)
			memstats[idx]++;
	}
}

void UsersReport::serialize_php(std::stringstream &out) {
	out << "a:5:{";
	out << "s:5:\"count\";i:" << count << ";";

	out << "s:8:\"memstats\";";
	out << "a:" << MEMSTATS_COUNT + 1 << ":{";
	out << "i:0;i:" << empty << ";";
	for (int i = 0; i < MEMSTATS_COUNT; i++)
		out << "i:" << (1 << (i + BUFFER_SCALE_MIN)) << ";i:" << memstats[i] << ";";
	out << "}";

	out << "s:13:\"deleted_count\";i:" << deleted_count << ";";

	out << "s:14:\"messages_count\";i:" << messages_count << ";";
	out << "s:9:\"msgsstats\";a:" << user_messages_size + 1 << ":{";
	for (int i = 0; i <= user_messages_size; i++)
		out << "i:" << i << ";i:" << msgstats[i] << ";";
	out << "}";

	out << "}";
}

UsersReport::UsersReport() {
	for (int i = 0; i < MEMSTATS_COUNT; i++)
		memstats[i] = 0;
	for (int i = 0; i <= user_messages_size; i++)
		msgstats[i] = 0;
	empty = 0;
	count = 0;
	messages_count = 0;
	deleted_count = 0;
}

