/*
 *  Copyright (C) 2009 Nicolas Vion <nico@picapo.net>
 *
 *   This file is part of Fluxy.
 *
 *   Fluxy is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Fluxy is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Fluxy; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef _USER_HH
#define _USER_HH

#include <cstdlib>
#include <stdint.h>
#include <sstream>
#include <string>

#include "fluxy.hh"
#include "pthread++.hh"
#include "result.hh"
#include "words_parser.hh"

struct UserMessageSearch {
	UserMessageId id;
	bool precise_type;
	UserMessageType type;

	void parse(WordsParser *parser);
	void output(std::stringstream &replication_query);
};

struct UserMessageValues {
	UserMessageId id;
	time_t date;
	UserMessageType type;
	UserMessageFlags flags;
	UserMessageVersion version;
	bool seen;
	std::string body;
};

struct UserMessageFilter {
	bool id;
	bool date;
	bool type;
	bool flags;
	bool version;
	bool seen;
	bool body;
};

struct UserMessageStorage {
	time_t date;
	UserMessageType type;
	UserMessageFlags flags;
	UserMessageVersion version;
	UserMessageId id;
	uint16_t body_len;
	char body;
} __attribute__ ((packed));

struct UserMessagesIndexItem {
	UserMessageStorage *message;
	size_t size_from_begining;
	size_t size;
	size_t body_len;
};

struct UserMessagesIndex {
	UserMessagesIndexItem items[USER_MESSAGES_SIZE];

	int make(void *buffer, int const count, time_t const until = 0);
	int find(int const count, UserMessageSearch const search);
	size_t size_from_begining(int const count);
};

class Fetcher;

struct UserMessages {
private:
	bool alloc(size_t const needed_size);

public:
	uint8_t buffer_count;
	BufferScale buffer_scale;
	char *buffer;

	void debug_buffer();
	bool add(UserMessageValues const values);
	bool del(UserMessageSearch const search);
	bool update(UserMessageSearch const search, UserMessageFilter const filter, UserMessageValues const values);
	bool seen(bool const all, UserMessageSearch const last);
	int count(bool const unseen_only = false);
	void serialize_php(std::stringstream &out);
	void cleanup();
	void clear();
	void fetch(UserId const user_id, Fetcher &fetcher, UserMessageFlags const mask = 0, UserMessageFlags const value = 0, time_t const until = 0, time_t const since = 0);
	bool parse_query(ClientResult &result, std::string const cmd_prefix, WordsParser *parser, std::stringstream &replication_query);
	void dump(FILE *f);
	UserId getid();
	UserScore score();
	UserMessages();
};

struct User {
public:
	UserId id;
	UserFlags flags;
	bool deleted;
	UserMessages messages;
	
	PMutex *lock();
	void serialize_php(std::stringstream &out);
	void show(std::stringstream &out);
	bool parse_query(ClientResult &result, std::string const cmd_prefix, WordsParser *parser);
	void cleanup();
	void clear();
	void dump(FILE *f);
	UserScore score();

	User(UserId const id);
	User(FILE *f);
};

#define MEMSTATS_COUNT (BUFFER_SCALE_MAX - BUFFER_SCALE_MIN + 1)
struct UsersReport {
private:
	int count;
	int messages_count;
	int memstats[MEMSTATS_COUNT];
	int msgstats[USER_MESSAGES_SIZE + 1];
	int empty;
	int deleted_count;

public:
	void add(User &user);
	void serialize_php(std::stringstream &out);
	UsersReport();
};

struct UserLocks {
	PMutex lock[1024];
};

#ifdef _USER_CC
UserLocks user_locks;
time_t user_messages_expiration;
int user_messages_size;
#else
extern UserLocks user_locks;
extern time_t user_messages_expiration;
extern int user_messages_size;
#endif


#endif
